package org.xbib.interlibrary.http.server.docs;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.CharBuffer;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.LinkedHashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.event.path.PathEventService;
import org.xbib.interlibrary.api.order.OrderSystem;
import org.xbib.interlibrary.db.DatabaseModule;
import org.xbib.interlibrary.docs.DocumentModule;
import org.xbib.interlibrary.docs.Layout;
import org.xbib.interlibrary.docs.OutputStyle;
import org.xbib.interlibrary.docs.form.PrintRLParser;
import org.xbib.interlibrary.http.server.InterlibraryServer;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.http.HttpAddress;
import org.xbib.net.http.HttpHeaderNames;
import org.xbib.net.http.HttpHeaderValues;
import org.xbib.net.http.HttpMethod;
import org.xbib.net.http.HttpResponseStatus;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.domain.BaseHttpDomainBuilder;
import org.xbib.net.http.server.domain.HttpDomain;
import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.route.BaseHttpRouterBuilder;
import org.xbib.net.http.server.service.BaseHttpService;
import org.xbib.settings.Settings;

public class HttpDocumentServer extends InterlibraryServer {

    private static final Logger logger = Logger.getLogger(HttpDocumentServer.class.getName());

    private final AtomicLong counter;

    public HttpDocumentServer() {
        super();
        this.counter = new AtomicLong();
    }

    @Override
    protected void customizeRouter(BaseHttpRouterBuilder baseHttpRouterBuilder) {
    }

    @Override
    protected HttpDomain buildDomain(HttpAddress httpAddress,
                                     HttpSecurityDomain securityDomain,
                                     BaseHttpDomainBuilder domainBuilder) throws IOException {
        DocumentModule documentModule = getModule(DocumentModule.class);
        String prefix = Objects.requireNonNull(getSettings().get("web.prefix"), "web.prefix must not be null");
        return domainBuilder
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/ping")
                    .setMethod(HttpMethod.GET, HttpMethod.POST, HttpMethod.HEAD)
                    .setHandler(ctx -> ctx.status(HttpResponseStatus.OK).body("OK"))
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/event/clock/suspend/{name}")
                    .setMethod(HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            String name = ctx.getRequest().getParameter().getAsString("name", Parameter.Domain.PATH);
                            documentModule.getEventManager().suspend(name);
                            ctx.status(HttpResponseStatus.OK).body("OK");
                        } catch (ParameterException e) {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/event/clock/resume/{name}/")
                    .setMethod(HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            String name = ctx.getRequest().getParameter().getAsString("name", Parameter.Domain.PATH);
                            documentModule.getEventManager().resume(name);
                            ctx.status(HttpResponseStatus.OK).body("OK");
                        } catch (ParameterException e) {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/form/{outputstyle}/{layout}/{ordernumber}")
                    .setMethod(HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            OutputStyle outputStyle = OutputStyle.valueOf(ctx.getRequest().getParameter()
                                    .getAsString("outputstyle", Parameter.Domain.PATH).toUpperCase(Locale.ROOT));
                            Layout layout = Layout.valueOf(ctx.getRequest().getParameter()
                                    .getAsString("layout", Parameter.Domain.PATH).toUpperCase(Locale.ROOT));
                            String orderNumber = ctx.getRequest().getParameter().getAsString("ordernumber", Parameter.Domain.PATH);
                            Boolean dryrun = ctx.getRequest().getParameter().getAsBoolean("dryrun", Parameter.Domain.QUERY);
                            dryrun = dryrun != null ? dryrun : false;
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            if (documentModule.form(orderNumber, layout, outputStyle, outputStream, dryrun)) {
                                byte[] bytes = outputStream.toByteArray();
                                if (bytes.length > 0) {
                                    ctx.status(HttpResponseStatus.OK)
                                            .setHeader(HttpHeaderNames.CONTENT_TYPE, "application/pdf")
                                            .setHeader(HttpHeaderNames.CONTENT_LENGTH, Long.toString(bytes.length))
                                            .body(ctx.getDataBufferFactory().wrap(bytes));
                                } else {
                                    ctx.status(HttpResponseStatus.OK)
                                            .setHeader(HttpHeaderNames.CONTENT_TYPE, "text/plain: charset=utf-8")
                                            .body("OK, but no form to show");
                                }
                            } else {
                                ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR);
                            }
                        } catch (ParameterException e) {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    }).build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/requestform/{ordersystem}/{ordernumber}")
                    .setMethod(HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            String orderSystemName = ctx.getRequest().getParameter().getAsString("ordersystem", Parameter.Domain.PATH);
                            OrderSystem orderSystem = OrderSystem.valueOf(orderSystemName.toUpperCase(Locale.ROOT));
                            String orderNumber = ctx.getRequest().getParameter().getAsString("ordernumber", Parameter.Domain.PATH);
                            documentModule.processRequestForm(orderSystem, orderNumber);
                            ctx.status(HttpResponseStatus.OK)
                                    .setHeader(HttpHeaderNames.CONTENT_TYPE, "text/plain: charset=utf-8")
                                    .body("OK");
                        } catch (IOException e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                            ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR);
                        } catch (ParameterException e) {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    }).build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/printrl")
                    .setMethod(HttpMethod.POST)
                    .setHandler(ctx -> {
                        try {
                            logger.log(Level.INFO, ctx.getRequest().asJson());
                            Boolean dryrun = ctx.getRequest().getParameter().getAsBoolean("dryrun", Parameter.Domain.QUERY);
                            dryrun = dryrun != null ? dryrun : false;
                            String contentType = ctx.getRequest().getHeaders().get(HttpHeaderNames.CONTENT_TYPE);
                            if (contentType != null && contentType.startsWith(HttpHeaderValues.TEXT_PLAIN)) {
                                CharBuffer charBuffer = ctx.getRequest().getBodyAsChars(StandardCharsets.UTF_8);
                                String printRL = charBuffer.toString();
                                logger.log(Level.INFO, printRL);
                                PrintRLParser printRLParser = new PrintRLParser(getConfiguration());
                                Map<String, Object> map = buildPrintRLMap();
                                printRLParser.parse(printRL, map);
                                ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                                if (documentModule.form(map, Layout.LEGACY, OutputStyle.SETTINGS, outputStream, dryrun)) {
                                    ctx.status(HttpResponseStatus.OK)
                                            .setHeader(HttpHeaderNames.CONTENT_TYPE, "text/plain")
                                            .body("OK");
                                } else {
                                    ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR);
                                }
                            }
                        } catch (ParameterException e) {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    }).build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/legacy")
                    .setMethod(HttpMethod.POST)
                    .setHandler(ctx -> {
                        // print legacy request form directly from json (internal)
                        String contentType = ctx.getRequest().getHeaders().get(HttpHeaderNames.CONTENT_TYPE);
                        if (contentType != null && contentType.startsWith("application/json")) {
                            Settings formSettings = Settings.settingsBuilder()
                                    .loadFromResource("json", ctx.getRequest().getInputStream())
                                    .build();
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            if (documentModule.getFormManager().getLegacyRequestForm().executeByDocumentEngine(formSettings, outputStream)) {
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(HttpHeaderNames.CONTENT_TYPE, "application/pdf")
                                        .setHeader(HttpHeaderNames.CONTENT_LENGTH, Long.toString(outputStream.size()))
                                        .body(ctx.getDataBufferFactory().wrap(outputStream.toByteArray()));
                            } else {
                                ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR);
                            }
                        } else {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    }).build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/medea")
                    .setMethod(HttpMethod.POST)
                    .setHandler(ctx -> {
                        // print medea request form directly from json (internal)
                        String contentType = ctx.getRequest().getHeaders().get(HttpHeaderNames.CONTENT_TYPE);
                        if (contentType != null && contentType.startsWith("application/json")) {
                            Settings formSettings = Settings.settingsBuilder()
                                    .loadFromResource("json", ctx.getRequest().getInputStream())
                                    .build();
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            if (documentModule.getFormManager().getMedeaRequestForm().executeByDocumentEngine(formSettings, outputStream)) {
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(HttpHeaderNames.CONTENT_TYPE, "application/pdf")
                                        .setHeader(HttpHeaderNames.CONTENT_LENGTH, Long.toString(outputStream.size()))
                                        .body(ctx.getDataBufferFactory().wrap(outputStream.toByteArray()));
                            } else {
                                ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR);
                            }
                        } else {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    }).build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/universal")
                    .setMethod(HttpMethod.POST)
                    .setHandler(ctx -> {
                        // print universal request form directly from json (internal)
                        String contentType = ctx.getRequest().getHeaders().get(HttpHeaderNames.CONTENT_TYPE);
                        if (contentType != null && contentType.startsWith("application/json")) {
                            Settings formSettings = Settings.settingsBuilder()
                                    .loadFromResource("json", ctx.getRequest().getInputStream())
                                    .build();
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            if (documentModule.getFormManager().getUniversalRequestForm().executeByDocumentEngine(formSettings, outputStream)) {
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(HttpHeaderNames.CONTENT_TYPE, "application/pdf")
                                        .setHeader(HttpHeaderNames.CONTENT_LENGTH, Long.toString(outputStream.size()))
                                        .body(ctx.getDataBufferFactory().wrap(outputStream.toByteArray()));
                            } else {
                                ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR);
                            }
                        } else {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    }).build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/usernotification")
                    .setMethod(HttpMethod.POST)
                    .setHandler(ctx -> {
                        // print user notification form directly from json (internal)
                        String contentType = ctx.getRequest().getHeaders().get(HttpHeaderNames.CONTENT_TYPE);
                        if (contentType != null && contentType.startsWith("application/json")) {
                            Settings formSettings = Settings.settingsBuilder()
                                    .loadFromResource("json", ctx.getRequest().getInputStream())
                                    .build();
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            if (documentModule.getFormManager().getUserNotificationForm().executeByDocumentEngine(formSettings, outputStream)) {
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(HttpHeaderNames.CONTENT_TYPE, "application/pdf")
                                        .setHeader(HttpHeaderNames.CONTENT_LENGTH, Long.toString(outputStream.size()))
                                        .body(ctx.getDataBufferFactory().wrap(outputStream.toByteArray()));
                            } else {
                                ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR);
                            }
                        } else {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    }).build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/convert")
                    .setMethod(HttpMethod.POST)
                    .setHandler(ctx -> {
                        // receive convert job in message body amd push it to a path-based queue
                        String contentType = ctx.getRequest().getHeaders().get(HttpHeaderNames.CONTENT_TYPE);
                        logger.log(Level.INFO, "convert: received content type = " + contentType);
                        if (contentType != null && contentType.startsWith("application/json")) {
                            String body = ctx.getRequest().getBodyAsChars(StandardCharsets.UTF_8).toString();
                            logger.log(Level.INFO, "convert: received body = " + body);
                            String key = "convert_" + LocalDateTime.now().toEpochSecond(ZoneOffset.UTC) +
                                    "_" + counter.incrementAndGet();
                            logger.log(Level.INFO, "convert: key = " + key);
                            PathEventService pathEventService = documentModule.getEventManager()
                                    .getPathEventManagerService().getPathEventService("convert");
                            if (pathEventService == null) {
                                logger.log(Level.WARNING, "convert: path event service 'convert' not configured, unable to publish convert request");
                                ctx.status(HttpResponseStatus.INTERNAL_SERVER_ERROR).done();
                            } else {
                                if (documentModule.getEventManager().getPathEventManagerService()
                                        .publishJson(pathEventService, key, ".json", body)) {
                                    logger.log(Level.INFO, "convert: publish of event succeeded");
                                    ctx.status(HttpResponseStatus.OK).done();
                                } else {
                                    ctx.status(HttpResponseStatus.CONFLICT).done();
                                }
                            }
                        } else {
                            ctx.status(HttpResponseStatus.BAD_REQUEST).done();
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/chart/{year}")
                    .setMethod(HttpMethod.POST)
                    .setHandler(ctx -> {
                        try {
                            String contentType = ctx.getRequest().getHeaders().get(HttpHeaderNames.CONTENT_TYPE);
                            if (contentType != null && contentType.startsWith("application/json")) {
                                Integer year = ctx.getRequest().getParameter().getAsInteger("year", Parameter.Domain.PATH);
                                String body = ctx.getRequest().getBodyAsChars(StandardCharsets.UTF_8).toString();
                                Map<String, Object> json = Json.toMap(body);
                                ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                                documentModule.chart(outputStream, year, json);
                                ctx.status(HttpResponseStatus.OK)
                                        .setHeader(HttpHeaderNames.CONTENT_TYPE, "image/svg+xml")
                                        .setHeader(HttpHeaderNames.CONTENT_DISPOSITION, "inline; filename=\"" + year + ".svg\"")
                                        .setHeader(HttpHeaderNames.CONTENT_LENGTH, Long.toString(outputStream.size()))
                                        .body(ctx.getDataBufferFactory().wrap(outputStream.toByteArray()));
                            } else {
                                ctx.status(HttpResponseStatus.BAD_REQUEST);
                            }
                        } catch (ParameterException e) {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/code39")
                    .setMethod(HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            String content = ctx.getRequest().getParameter().getAsString("content", Parameter.Domain.QUERY);
                            if (content == null || content.isEmpty()) {
                                ctx.status(HttpResponseStatus.BAD_REQUEST);
                                return;
                            }
                            Integer width = ctx.getRequest().getParameter().getAsInteger("width", Parameter.Domain.QUERY);
                            if (width == null) {
                                width = 400;
                            }
                            Integer height = ctx.getRequest().getParameter().getAsInteger("height", Parameter.Domain.QUERY);
                            if (height == null) {
                                height = 100;
                            }
                            logger.log(Level.INFO, "code39: content = " + content + " width = " + width + " height = " + height);
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            documentModule.code39(outputStream, "png", content, width, height);
                            ctx.status(HttpResponseStatus.OK)
                                    .setHeader(HttpHeaderNames.CONTENT_TYPE, "image/png")
                                    .setHeader(HttpHeaderNames.CONTENT_DISPOSITION, "inline; filename=\"" + content + ".png\"")
                                    .setHeader(HttpHeaderNames.CONTENT_LENGTH, Long.toString(outputStream.size()))
                                    .body(ctx.getDataBufferFactory().wrap(outputStream.toByteArray()));
                        } catch (ParameterException e) {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    })
                    .build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/qrcode")
                    .setMethod(HttpMethod.GET)
                    .setHandler(ctx -> {
                        try {
                            String content = ctx.getRequest().getParameter().getAsString("content", Parameter.Domain.QUERY);
                            if (content == null || content.isEmpty()) {
                                ctx.status(HttpResponseStatus.BAD_REQUEST);
                                return;
                            }
                            Integer width = ctx.getRequest().getParameter().getAsInteger("width", Parameter.Domain.QUERY);
                            if (width == null) {
                                width = 200;
                            }
                            Integer height = ctx.getRequest().getParameter().getAsInteger("height", Parameter.Domain.QUERY);
                            if (height == null) {
                                height = 200;
                            }
                            logger.log(Level.INFO, "qrcode: content = " + content + " width = " + width + " height = " + height);
                            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
                            documentModule.qrcode(outputStream, "png", content, width, height);
                            ctx.status(HttpResponseStatus.OK)
                                    .setHeader(HttpHeaderNames.CONTENT_TYPE, "image/png")
                                    .setHeader(HttpHeaderNames.CONTENT_DISPOSITION, "inline; filename=\"" + content + ".png\"")
                                    .setHeader(HttpHeaderNames.CONTENT_LENGTH, Long.toString(outputStream.size()))
                                    .body(ctx.getDataBufferFactory().wrap(outputStream.toByteArray()));
                        } catch (ParameterException e) {
                            ctx.status(HttpResponseStatus.BAD_REQUEST);
                        }
                    }).build())
            .addService(BaseHttpService.builder()
                    .setPrefix(prefix)
                    .setPath("/pull")
                    .setMethod(HttpMethod.GET)
                    .setHandler(ctx -> {
                        documentModule.pull();
                        ctx.status(HttpResponseStatus.OK).body("OK");
                    })
                    .build())
            .build();
    }

    @Override
    protected void customizeApplication(Application application) {
        DatabaseModule databaseModule = getModule(DatabaseModule.class);
        application.getAttributes().put("databasemodule", databaseModule);
        application.addModule(new HttpDocumentServerModule(application, "app", getSettings()));
    }

    private static Map<String, Object> buildPrintRLMap() {
        Map<String, Object> map = new LinkedHashMap<>();
        map.put("font", "sourcesans, notosans, notosanscjksc");
        map.put("cut", "true");
        map.put("perforation", "true");
        map.put("outline", "true");
        return map;
    }
}
